
local k = {
	0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1,
	0x923f82a4, 0xab1c5ed5, 0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
	0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174, 0xe49b69c1, 0xefbe4786,
	0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
	0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147,
	0x06ca6351, 0x14292967, 0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
	0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85, 0xa2bfe8a1, 0xa81a664b,
	0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
	0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a,
	0x5b9cca4f, 0x682e6ff3, 0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
	0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2,
}

local w = { }

function Sha256(message)
	local function ror(x, y)
		return bit.rshift(x, y) + bit.lshift(x, 32 - y)
	end
	
	local h0 = 0x6a09e667
	local h1 = 0xbb67ae85
	local h2 = 0x3c6ef372
	local h3 = 0xa54ff53a
	local h4 = 0x510e527f
	local h5 = 0x9b05688c
	local h6 = 0x1f83d9ab
	local h7 = 0x5be0cd19
	
	local msg = message .. "\128" .. ("\0"):rep(64 - (#message + 8) % 64 + 3) ..
		string.char(
			bit.rshift(#message * 8, 24),
			bit.rshift(#message * 8, 16) % 0x100,
			bit.rshift(#message * 8, 8)  % 0x100,
			(#message * 8) % 0x100
		)
	
	for chunkid = 1, #msg / 64 do
		for i = 1, 16 do
			local location   = (i - 1) * 4 + (chunkid - 1) * 64
			local A, B, C, D = msg:byte(location + 1, location + 4)
			
			w[i] = bit.lshift(A, 24) + bit.lshift(B, 16) + bit.lshift(C, 8) + D
		end
		
		for i = 17, 64 do
			local s0 = bit.bxor(ror(w[i - 15], 7), ror(w[i - 15], 18), bit.rshift(w[i - 15], 3))
			local s1 = bit.bxor(ror(w[i - 2], 17), ror(w[i - 2],  19), bit.rshift(w[i - 2], 10))
			
			w[i] = (w[i - 16] + s0 + w[i - 7] + s1) % 2 ^ 32
		end
		
		local a, b, c, d, e, f, g, h = h0, h1, h2, h3, h4, h5, h6, h7
		
		for i = 1, 64 do
			local ma = bit.bxor(bit.band(a, b), bit.band(a, c), bit.band(b, c))
			local ch = bit.bxor(bit.band(e, f), bit.band(bit.bnot(e), g))
			local s0 = bit.bxor(ror(a, 2), ror(a, 13), ror(a, 22))
			local s1 = bit.bxor(ror(e, 6), ror(e, 11), ror(e, 25))
			local t1 = h + s1 + ch + k[i] + w[i]
			local t2 = s0 + ma
			
			h = g
			g = f
			f = e
			e = (d + t1) % 2 ^ 32
			d = c
			c = b
			b = a
			a = (t1 + t2) % 2 ^ 32
		end
		
		h0 = (h0 + a) % 2 ^ 32
		h1 = (h1 + b) % 2 ^ 32
		h2 = (h2 + c) % 2 ^ 32
		h3 = (h3 + d) % 2 ^ 32
		h4 = (h4 + e) % 2 ^ 32
		h5 = (h5 + f) % 2 ^ 32
		h6 = (h6 + g) % 2 ^ 32
		h7 = (h7 + h) % 2 ^ 32
	end
	
	return ("%08x%08x%08x%08x%08x%08x%08x%08x"):format(h0, h1, h2, h3, h4, h5, h6, h7)
end

